@@ -1306,7 +1306,21 @@ test_unix_fd_source (void)
 
   out = in = FALSE;
   out_source = g_unix_fd_source_new (fds[1], G_IO_OUT);
-  g_source_set_callback (out_source, (GSourceFunc) flag_bool, &out, NULL);
+  /* -Wcast-function-type complains about casting 'flag_bool' to GSourceFunc.
+   * GCC has no way of knowing that it will be cast back to GUnixFDSourceFunc
+   * before being called. Although GLib itself is not compiled with
+   * -Wcast-function-type, applications that use GLib may well be (since
+   * -Wextra includes it), so we provide a G_SOURCE_FUNC() macro to suppress
+   * the warning. We check that it works here.
+   */
+#if G_GNUC_CHECK_VERSION(8, 0)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic error "-Wcast-function-type"
+#endif
+  g_source_set_callback (out_source, G_SOURCE_FUNC (flag_bool), &out, NULL);
+#if G_GNUC_CHECK_VERSION(8, 0)
+#pragma GCC diagnostic pop
+#endif
   g_source_attach (out_source, NULL);
   assert_main_context_state (1,
                              fds[1], G_IO_OUT, 0);
